# -*- coding: utf-8 -*-
import os
import abc
import json
import time
from buffalo.misc import aux, log
from buffalo.misc.log import pbar

from hyperopt import hp, fmin, tpe, Trials


class Optimizable(object):
    def __init__(self, *args, **kwargs):
        self._optimization_info = {'trials': Trials(), 'best': {}}
        self._temporary_opt_file = aux.get_temporary_file()
        self.optimize_after_callback_fn = kwargs.get('optimize_after_callback_fn')

    def _get_space(self, options):
        spaces = {}
        for param_name, data in options.items():
            algo, opt = data
            if algo == 'randint':
                spaces[param_name] = opt[1] + hp.randint(opt[0], opt[2] - opt[1])
            else:
                spaces[param_name] = getattr(hp, algo)(*opt)
        return spaces

    @abc.abstractmethod
    def _optimize(self):
        raise NotImplemented

    def optimize(self):
        opt = self.opt.optimize
        iters, max_trials = 0, opt.get('max_trials', -1)
        space = self._get_space(opt.space)
        with log.pbar(log.INFO, desc='optimizing... ',
                      total=None if max_trials == -1 else max_trials,
                      mininterval=30) as pbar:
            tb_opt = None
            tb_opt, self.opt.tensorboard = self.opt.tensorboard, tb_opt  # trick
            if opt.start_with_default_parameters:
                with log.supress_log_level(log.WARN):
                    loss = self._optimize({})
                self.logger.info(f'Starting with default parameter result: {loss}')
                self._optimization_info['best'] = loss
                if opt.deployment:
                    self.logger.info('Saving model... to {}'.format(self.opt.model_path))
                    self.save(self.opt.model_path)
            # NOTE: need better way
            tb_opt, self.opt.tensorboard = self.opt.tensorboard, tb_opt  # trick
            self.initialize_tensorboard(1000000 if max_trials == -1 else max_trials,
                                        name_postfix='.optimize')
            tb_opt, self.opt.tensorboard = self.opt.tensorboard, tb_opt  # trick
            while(max_trials):
                with log.supress_log_level(log.WARN):
                    best = fmin(fn=self._optimize,
                                space=space,
                                algo=tpe.suggest,
                                max_evals=len(self._optimization_info['trials'].trials) + 1,
                                trials=self._optimization_info['trials'],
                                show_progressbar=False)
                tb_opt, self.opt.tensorboard = self.opt.tensorboard, tb_opt  # trick
                self.update_tensorboard_data(self._optimize_loss)
                tb_opt, self.opt.tensorboard = self.opt.tensorboard, tb_opt  # trick
                iters += 1
                max_trials -= 1
                if self._optimization_info.get('best', {}).get('loss', 987654321) > self._optimize_loss['loss']:
                    is_first_time = self._optimization_info['best'] == {}
                    best = self._optimize_loss  # we cannot use return value of hyperopt due to randint behavior patch
                    self.logger.info(f'Found new best parameters: {best} @ iter {iters}')
                    self._optimization_info['best'] = best
                    if opt.deployment and (is_first_time or not opt.min_trials or opt.min_trials >= iters):
                        if not self.opt.model_path:
                            raise RuntimeError('Failed to dump model: model path is not defined')
                        self.logger.info('Saving model... to {}'.format(self.opt.model_path))
                        self.save(self.opt.model_path)
                if self.optimize_after_callback_fn:
                    self.optimize_after_callback_fn(self)
                pbar.update(1)
                self.logger.debug('Params({}) Losses({})'.format(self._optimize_params, self._optimize_loss))
            tb_opt, self.opt.tensorboard = self.opt.tensorboard, tb_opt  # trick
            self.finalize_tensorboard()

    def get_optimization_data(self):
        return self._optimization_info



from n2 import HnswIndex
import random

f = 3
t = HnswIndex(f, "L2")  # HnswIndex(f, "L2 or angular")
for i in xrange(1000):
    v = [random.gauss(0, 1) for z in xrange(f)]
    t.add_data(v)

t.build(m=5, max_m0=10, n_threads=4)
t.save('test.n2')

u = HnswIndex(f, "L2")
u.load('test.n2')

search_id = 1
k = 3
neighbor_ids = u.search_by_id(search_id, k)
print(
    "[search_by_id]: Nearest neighborhoods of id {}: {}".format(
        search_id,
        neighbor_ids))

example_vector_query = [random.gauss(0, 1) for z in xrange(f)]
nns = u.search_by_vector(example_vector_query, k, include_distances=True)
print(
    "[search_by_vector]: Nearest neighborhoods of vector {}: {}".format(
        example_vector_query,
        nns))



# -*- coding: utf-8 -*-
import os
import psutil
import warnings
import traceback
from collections import Counter

import h5py
import numpy as np

from buffalo.data import prepro
from buffalo.misc import aux, log
from buffalo.data.base import Data, DataOption


class StreamOptions(DataOption):
    """
    The option class for Stream class
    Options:
        type: Must be "stream"
        input:
            main: Path to main stream file.
            uid: User names corresponding to each line.
            iid: Item names corresponding to each item id.
        data:
            validation: See validation section.
            batch_mb: Internal batch size. Generally, the larger size, faster.
            use_cache: Set True to use already built data, otherwise building new one every time.
            tmp_dir: Where temporary files goes on.
            path: Output path of Stream.
            internal_data_type: "stream" or "matrix"
                stream: Data file treated as is(i.e. streaming data)
                matrix: Translate data file into sparse matrix format(it lose sequential information, but more compact for data size).
            sppmi:
                windows: window size to set relation between word and context
                k: sppmi shift k (serves as negative sampling size in w2v model)
    """
    def get_default_option(self) -> aux.Option:
        opt = {
            'type': 'stream',
            'input': {
                'main': '',
                'uid': '',  # if not set, row-id is used as userid.
                'iid': ''  # if not set, col-id is used as userid.
            },
            'data': {
                'validation': {
                    'name': 'newest',  # sample or oldest or newest
                    'p': 0.01,  # if set oldest or newest, ignored
                    'n': 1,  # if set sample, ignored
                    'max_samples': 500
                },
                'sppmi': {
                    # 'windows': 5,
                    # 'k': 1
                },
                'batch_mb': 1024,
                'use_cache': False,
                'tmp_dir': '/tmp/',
                'path': './stream.h5py',
                'internal_data_type': 'stream'  # if set to 'matrix', internal data stored as like matrix market format
            }
        }
        return aux.Option(opt)

    def is_valid_option(self, opt) -> bool:
        assert super(StreamOptions, self).is_valid_option(opt)
        if not opt['type'] == 'stream':
            raise RuntimeError('Invalid data type: %s' % opt['type'])
        return True


class Stream(Data):
    def __init__(self, opt, *args, **kwargs):
        super(Stream, self).__init__(opt, *args, **kwargs)
        self.name = 'Stream'
        self.logger = log.get_logger('Stream')
        self.data_type = 'stream'

    def _create(self, data_path, P):
        def get_max_column_length(fname):
            with open(fname) as fin:
                max_col = 0
                for l in fin:
                    max_col = max(max_col, len(l))
            return max_col
        uid_path, iid_path, main_path = P['uid_path'], P['iid_path'], P['main_path']
        if uid_path:
            with open(uid_path) as fin:
                num_users = len([1 for _ in fin])
        else:
            with open(main_path) as fin:
                num_users = len([1 for _ in fin])

        uid_max_col = len(str(num_users)) + 1
        if uid_path:
            uid_max_col = get_max_column_length(uid_path) + 1

        vali_n = self.opt.data.validation.get('n', 0)
        num_nnz, vali_limit, itemids = 0, 0, set()
        self.logger.info(f'gathering itemids from {main_path}...')
        if self.opt.data.validation.name not in ["oldest", "newest"]:
            vali_n = 0
        with open(main_path) as fin:
            for line in log.iter_pbar(log_level=log.DEBUG, iterable=fin):
                data = line.strip().split()
                if not iid_path:
                    itemids |= set(data)

                data_size = len(data)
                _vali_size = min(vali_n, len(data) - 1)
                vali_limit += _vali_size
                if self.opt.data.internal_data_type == 'stream':
                    num_nnz += (data_size - _vali_size)
                elif self.opt.data.internal_data_type == 'matrix':
                    num_nnz += len(set(data[:(data_size - _vali_size)]))
        if iid_path:
            with open(iid_path) as fin:
                itemids = {iid.strip(): idx + 1 for idx, iid in enumerate(fin)}
        else:  # in case of item information is not given
            itemids = {i: idx + 1 for idx, i in enumerate(itemids)}
        iid_max_col = max(len(k) + 1 for k in itemids.keys())
        num_items = len(itemids)

        self.logger.info('Found %d unique itemids' % len(itemids))

        try:
            db = self._create_database(data_path,
                                       num_users=num_users,
                                       num_items=num_items,
                                       num_nnz=num_nnz,
                                       uid_max_col=uid_max_col,
                                       iid_max_col=iid_max_col,
                                       num_validation_samples=vali_limit)
            idmap = db['idmap']
            # if not given, assume id as is
            if uid_path:
                with open(uid_path) as fin:
                    idmap['rows'][:] = np.loadtxt(fin, dtype=f'S{uid_max_col}')
            else:
                idmap['rows'][:] = np.array([str(i) for i in range(1, num_users + 1)],
                                            dtype=f'S{uid_max_col}')
            if iid_path:
                with open(iid_path) as fin:
                    idmap['cols'][:] = np.loadtxt(fin, dtype=f'S{iid_max_col}')
            else:
                cols = sorted(itemids.items(), key=lambda x: x[1])
                cols = [k for k, _ in cols]
                idmap['cols'][:] = np.array(cols, dtype=f'S{iid_max_col}')
        except Exception as e:
            self.logger.error('Cannot create db: %s' % (str(e)))
            self.logger.error(traceback.format_exc())
            raise
        return db, itemids

    def _build_data(self, db, working_data_path, validation_data):
        if self.opt.data.internal_data_type == 'stream':
            super()._build_data(db, working_data_path, validation_data,
                                target_groups=['rowwise'],
                                sort=False)  # keep order
        elif self.opt.data.internal_data_type == 'matrix':
            super()._build_data(db, working_data_path, validation_data,
                                target_groups=['rowwise', 'colwise'])

    def _build_sppmi(self, db, working_data_path, k):
        self.logger.info(f"build sppmi (shift k: {k})")
        appearances = {}
        nnz = 0
        aux.psort(working_data_path, key=1)
        with open(working_data_path, "r") as fin, \
            open(aux.get_temporary_file(root=self.opt.data.tmp_dir), "w") as w:
            D = sum(1 for line in fin)
            fin.seek(0)
            probe, chunk = "junk", []
            for line in fin:
                _w, _c = line.strip().split()
                if probe != _w:
                    appearances[probe] = len(chunk)
                    for __c, cnt in Counter(chunk).items():
                        if int(probe) < int(__c):
                            continue
                        pmi = np.log(cnt) + np.log(D) - \
                            np.log(appearances[probe]) - np.log(appearances[__c])
                        sppmi = pmi - np.log(k)
                        if sppmi > 0:
                            w.write(f"{probe} {__c} {sppmi}\n")
                            w.write(f"{__c} {probe} {sppmi}\n")
                            nnz += 2
                    probe, chunk = _w, []
                chunk.append(_c)
        aux.psort(w.name)
        self.logger.info(f"convert from {working_data_path} to {w.name}")
        db.create_group("sppmi")
        db.attrs["sppmi_nnz"] = nnz
        self.logger.info(f"sppmi nnz: {nnz}")
        sz = db.attrs["num_items"]
        db["sppmi"].create_dataset("indptr", (sz,), dtype='int64', maxshape=(sz,))
        db["sppmi"].create_dataset("key", (nnz,), dtype='int32', maxshape=(nnz,))
        db["sppmi"].create_dataset("val", (nnz,), dtype='float32', maxshape=(nnz,))
        self.logger.info('Disk-based Compressing...')
        job_files = self._chunking_into_bins(w.name, nnz, sz, 0)
        self._build_compressed_triplets(db["sppmi"],
                                        job_files,
                                        num_lines=nnz,
                                        max_key=sz,
                                        is_colwise=0)

    def _create_working_data(self, db, stream_main_path, itemids,
                             with_sppmi=False, windows=5):
        vali_method = None if 'vali' not in db else db['vali'].attrs['method']
        vali_indexes, vali_n = set(), 0
        if vali_method == 'sample':
            vali_indexes = set(db['vali']['indexes'])
        elif vali_method in ['newest']:
            vali_n = db['vali'].attrs['n']
        vali_lines = []
        users = db['idmap']['rows'][:]

        with warnings.catch_warnings():
            warnings.simplefilter("ignore", ResourceWarning)
            if with_sppmi:
                w_sppmi = open(aux.get_temporary_file(root=self.opt.data.tmp_dir), "w")
            file_path = aux.get_temporary_file(root=self.opt.data.tmp_dir)
            with open(stream_main_path) as fin,\
                open(file_path, 'w') as w:
                total_index = 0
                internal_data_type = self.opt.data.internal_data_type
                for line_idx, data in log.iter_pbar(log_level=log.DEBUG, iterable=enumerate(fin)):
                    data = data.strip().split()
                    total_data_size = len(data)
                    user = line_idx + 1
                    vali_data, train_data = [], []
                    if vali_method in ['newest']:
                        vali_data_size = min(vali_n, len(data) - 1)
                        train_data_size = len(data) - vali_data_size
                        vali = data[train_data_size:]
                        data = data[:train_data_size]
                        for col, val in Counter(vali).items():
                            col = itemids[col]
                            vali_data.append(col)
                    if internal_data_type == 'stream':
                        for idx, col in enumerate(data):
                            col = itemids[col]
                            if (idx + total_index) in vali_indexes:
                                vali_data.append(col)
                            else:
                                train_data.append(col)
                    elif internal_data_type == 'matrix':
                        for idx, col in enumerate(data):
                            col = itemids[col]
                            if (idx + total_index) in vali_indexes:
                                vali_data.append(col)
                            else:
                                train_data.append(col)
                    total_index += len(data)
                    if internal_data_type == 'stream':
                        for col in train_data:
                            w.write(f'{user} {col} 1\n')
                        for col in vali_data:
                            vali_lines.append(f'{user} {col} {val}')
                    else:
                        for col, val in Counter(train_data).items():
                            w.write(f'{user} {col} {val}\n')
                        for col, val in Counter(vali_data).items():
                            vali_lines.append(f'{user} {col} {val}')
                    if with_sppmi:
                       sz = len(train_data)
                       for i in range(sz):
                           beg, end = i + 1, i + windows + 1
                           for j in range(beg, end):
                               if j >= sz:
                                   break
                               _w, _c = train_data[i], train_data[j]
                               w_sppmi.write(f'{_w} {_c}\n')
                               w_sppmi.write(f'{_c} {_w}\n')
                if with_sppmi:
                    w_sppmi.close()
                    return w.name, vali_lines, w_sppmi.name
                return w.name, vali_lines, None

    def create(self) -> h5py.File:
        stream_main_path = self.opt.input.main
        stream_uid_path = self.opt.input.uid
        stream_iid_path = self.opt.input.iid

        data_path = self.opt.data.path
        if os.path.isfile(data_path) and self.opt.data.use_cache:
            self.logger.info('Use cached DB on %s' % data_path)
            self.open(data_path)
            return

        self.logger.info('Create database from stream data')
        with open(stream_main_path) as fin:
            self.logger.debug('Building meta part...')
            db, itemids = self._create(data_path,
                                       {'main_path': stream_main_path,
                                        'uid_path': stream_uid_path,
                                        'iid_path': stream_iid_path})
            sppmi_opt = self.opt.data.sppmi
            try:
                self.logger.info('Creating working data...')
                if sppmi_opt:
                    tmp_main, validation_data, tmp_sppmi = \
                        self._create_working_data(db, stream_main_path, itemids, True, sppmi_opt.windows)
                else:
                    tmp_main, validation_data, tmp_sppmi = \
                        self._create_working_data(db, stream_main_path, itemids, False)
                self.logger.debug(f'Working data is created on {tmp_main}')
                self.logger.info('Building data part...')
                self._build_data(db, tmp_main, validation_data)
                if sppmi_opt:
                    self.logger.debug(f'sppmi data is created on {tmp_sppmi}')
                    self._build_sppmi(db, tmp_sppmi, sppmi_opt.k)
                db.attrs['completed'] = 1
                db.close()
                self.handle = h5py.File(data_path, 'r')
                self.path = data_path
            except Exception as e:
                self.logger.error('Cannot create db: %s' % (str(e)))
                self.logger.error(traceback.format_exc())
                raise
            finally:
                if hasattr(self, 'patr'):
                    if os.path.isfile(self.path):
                        os.remove(self.path)
        self.logger.info('DB built on %s' % data_path)



# -*- coding: utf-8 -*-
import math

import numpy as np
from buffalo.parallel._core import quickselect


class Evaluable(object):
    def __init__(self, *args, **kargs):
        pass

    def prepare_evaluation(self):
        if not self.opt.validation or not self.data.has_group('vali'):
            return
        vali = self.data.get_group('vali')
        validation_seen = {}
        rows = set([r for r in vali['row'][::]])
        max_seen_size = 0
        for rowid in rows:
            seen, *_ = self.data.get(rowid)
            validation_seen[rowid] = set(seen)
            max_seen_size = max(len(seen), max_seen_size)
        self._validation_seen = validation_seen
        self._validation_max_seen_size = max_seen_size

    def show_validation_results(self):
        results = self.get_validation_results()
        if not results:
            return 'No validation results'
        return 'Validation results: ' + ', '.join(f'{k}: {v:0.5f}' for k, v in results.items())

    def get_validation_results(self):
        if not self.opt.validation or not self.data.has_group('vali'):
            return

        results = {}
        results.update(self._evaluate_ranking_metrics())
        results.update(self._evaluate_score_metrics())
        return results

    def get_topk(self, scores, k, sorted=True, num_threads=4):
        # NOTE: Is it necessary condition?
        # assert k < scores.shape[1], f"k ({k}) should be smaller than cols ({scores.shape[1]})"
        is_many = True
        if len(scores.shape) == 1:
            scores = scores.reshape(1, scores.shape[0])
            is_many = False
        k = min(k, scores.shape[1])
        assert k > 0, f'k({k}) or cols({scores.shape[1]}) should be greater than 0'
        result = np.empty(shape=(scores.shape[0], k), dtype=np.int32)
        quickselect(scores, result, sorted, num_threads)
        return result if is_many else result[0]

    def _evaluate_ranking_metrics(self):
        batch_size = self.opt.validation.get('batch', 128)
        topk = self.opt.validation.topk
        vali = self.data.get_group('vali')
        row = vali['row'][::]
        col = vali['col'][::]
        rows, gt = set(), {}
        for r, c in zip(row, col):
            gt.setdefault(r, set()).add(c)
            rows.add(r)
        rows = list(rows)

        NDCG = 0.0
        AP = 0.0
        HIT = 0.0
        N = 0.0
        for index in range(0, len(rows), batch_size):
            recs = self._get_topk_recommendation(rows[index:index + batch_size],
                                                 topk=topk + self._validation_max_seen_size)
            for row, _topk in recs:
                seen = self._validation_seen.get(row, set())
                _topk = [t for t in _topk if t not in seen][:topk]
                _gt = gt[row]

                # accuracy
                hit = len(set(_topk) & _gt) / len(_gt)
                HIT += hit

                # ndcg, map
                idcg = sum([1.0 / math.log(i + 2, 2)
                            for i in range(min(len(_gt), len(_topk)))])
                dcg = 0.0
                hit, ap = 0.0, 0.0
                for i, r in enumerate(_topk):
                    if r in _gt:
                        hit += 1
                        ap += (hit / (i + 1.0))
                    if r not in _gt:
                        continue
                    rank = i + 1
                    dcg += 1.0 / math.log(rank + 1, 2)
                ndcg = dcg / idcg
                NDCG += ndcg
                ap /= min(len(_gt), len(_topk))
                AP += ap
                N += 1.0
        NDCG /= N
        AP /= N
        ACC = HIT / N
        ret = {'ndcg': NDCG, 'map': AP, 'accuracy': ACC}
        return ret

    def _evaluate_score_metrics(self):
        vali = self.data.get_group('vali')
        row = vali['row'][::]
        col = vali['col'][::]
        val = vali['val'][::]
        row_col_pairs = zip(row, col)
        vals = {(r, c): v for r, c, v in zip(row, col, val)}
        scores = self.get_scores(row_col_pairs)
        ERROR = 0.0
        RMSE = 0.0
        for (r, c), p in scores.items():
            err = p - vals[(r, c)]
            ERROR += abs(err)
            RMSE += err * err
        RMSE /= len(scores)
        RMSE = RMSE ** 0.5
        ERROR /= len(scores)
        return {'rmse': RMSE, 'error': ERROR}



# -*- coding: utf-8 -*-
import abc

import numpy as np

from buffalo.algo.als import ALS
from buffalo.algo.cfr import CFR
from buffalo.algo.w2v import W2V
from buffalo.algo.bpr import BPRMF

from buffalo.parallel._core import dot_topn, ann_search


class Parallel(abc.ABC):
    def __init__(self, algo, *argv, **kwargs):
        super().__init__()
        if not isinstance(algo, (ALS, CFR, W2V, BPRMF)):
            raise ValueError('Not supported algo type: %s' % type(algo))
        self.algo = algo
        self.num_workers = int(kwargs['num_workers'])
        self._ann_list = {}

    def _most_similar(self, group, indexes, Factor, topk, pool, ef_search, use_mmap):
        dummy_bias = np.array([[]], dtype=np.float32)
        out_keys = np.zeros(shape=(len(indexes), topk), dtype=np.int32)
        out_scores = np.zeros(shape=(len(indexes), topk), dtype=np.float32)
        if group in self._ann_list:
            if ef_search == -1:
                ef_search = topk * 10
            ann_search(self._ann_list[group].encode('utf8'), ef_search, use_mmap, indexes, Factor, Factor, dummy_bias, out_keys, out_scores, pool, topk, self.num_workers)
        else:
            dot_topn(indexes, Factor, Factor, dummy_bias, out_keys, out_scores, pool, topk, self.num_workers)
        return out_keys, out_scores

    def set_hnsw_index(self, path, group):
        """ Set N2 HNSW Index for apporximate nearest neighbor features.

        N2 is a open source project for approximate nearest neighbor.
        For more detail please see the https://github.com/kakao/n2.

        :param str path: The path of hnsw index file.
        :param str group: Indexed data group.
        """
        self._ann_list[group] = path

    @abc.abstractmethod
    def most_similar(self, keys, topk=10, group='item', pool=None, repr=False, ef_search=-1, use_mmap=True):
        """Caculate TopK most similar items for each keys in parallel processing.

        :param list keys: Query Keys
        :param int topk: Number of topK
        :param str group: Data group where to find (default: item)
        :param pool: The list of item keys to find for.
            If it is a numpy.ndarray instance then it treat as index of items and it would be helpful for calculation speed. (default: None)
        :type pool: list or numpy.ndarray
        :param bool repr: Set True, to return as item key instead index.
        :param int ef_search: This parameter is passed to N2 when hnsw_index was given for the group. (default: -1 which means topk * 10)
        :param use_mmap: This parameter is passed to N2 when hnsw_index given for the group. (default: True)
        :return: list of tuple(key, score)
        """
        raise NotImplemented

    def _topk_recommendation(self, indexes, FactorP, FactorQ, topk, pool):
        dummy_bias = np.array([[]], dtype=np.float32)
        out_keys = np.zeros(shape=(len(indexes), topk), dtype=np.int32)
        out_scores = np.zeros(shape=(len(indexes), topk), dtype=np.float32)
        dot_topn(indexes, FactorP, FactorQ, dummy_bias, out_keys, out_scores, pool, topk, self.num_workers)
        return out_keys, out_scores

    def _topk_recommendation_bias(self, indexes, FactorP, FactorQ, FactorQb, topk, pool):
        out_keys = np.zeros(shape=(len(indexes), topk), dtype=np.int32)
        out_scores = np.zeros(shape=(len(indexes), topk), dtype=np.float32)
        dot_topn(indexes, FactorP, FactorQ, FactorQb, out_keys, out_scores, pool, topk, self.num_workers)
        return out_keys, out_scores

    @abc.abstractmethod
    def topk_recommendation(self, keys, topk=10, pool=None, repr=False):
        """Caculate TopK recommendation for each users in parallel processing.

        :param list keys: Query Keys
        :param int topk: Number of topK
        :param bool repr: Set True, to return as item key instead index.
        :return: list of tuple(key, score)
        """
        raise NotImplemented


class ParALS(Parallel):
    def __init__(self, algo, **kwargs):
        num_workers = int(kwargs.get('num_workers', algo.opt.num_workers))
        super().__init__(algo, num_workers=num_workers)

    def most_similar(self, keys, topk=10, group='item', pool=None, repr=False, ef_search=-1, use_mmap=True):
        """See the documentation of Parallel."""
        self.algo.normalize(group=group)
        indexes = self.algo.get_index_pool(keys, group=group)
        keys = [k for k, i in zip(keys, indexes) if i is not None]
        indexes = np.array([i for i in indexes if i is not None], dtype=np.int32)
        if pool is not None:
            pool = self.algo.get_index_pool(pool, group=group)
            if len(pool) == 0:
                raise RuntimeError('pool is empty')
        else:
            # It assume that empty pool menas for all items
            pool = np.array([], dtype=np.int32)
        if group == 'item':
            topks, scores = super()._most_similar(group, indexes, self.algo.Q, topk, pool, ef_search, use_mmap)
            if repr:
                topks = [[self.algo._idmanager.itemids[t] for t in tt if t != -1] for tt in topks]
            return topks, scores
        elif group == 'user':
            topks, scores = super()._most_similar(group, indexes, self.algo.P, topk, pool, ef_search, use_mmap)
            if repr:
                topks = [[self.algo._idmanager.userids[t] for t in tt if t != -1] for tt in topks]
            return topks, scores
        raise ValueError(f'Not supported group: {group}')

    def topk_recommendation(self, keys, topk=10, pool=None, repr=False):
        """See the documentation of Parallel."""
        if self.algo.opt._nrz_P or self.algo.opt._nrz_Q:
            raise RuntimeError('Cannot make topk recommendation with normalized factors')
        # It is possible to skip make recommendation for not-existed keys.
        indexes = self.algo.get_index_pool(keys, group='user')
        keys = [k for k, i in zip(keys, indexes) if i is not None]
        indexes = np.array([i for i in indexes if i is not None], dtype=np.int32)
        if pool is not None:
            pool = self.algo.get_index_pool(pool, group='item')
            if len(pool) == 0:
                raise RuntimeError('pool is empty')
        else:
            # It assume that empty pool menas for all items
            pool = np.array([], dtype=np.int32)
        topks, scores = super()._topk_recommendation(indexes, self.algo.P, self.algo.Q, topk, pool)
        if repr:
            mo = np.int32(-1)
            topks = [[self.algo._idmanager.itemids[t] for t in tt if t != mo]
                     for tt in topks]
        return keys, topks, scores


class ParBPRMF(ParALS):
    def topk_recommendation(self, keys, topk=10, pool=None, repr=False):
        """See the documentation of Parallel."""
        if self.algo.opt._nrz_P or self.algo.opt._nrz_Q:
            raise RuntimeError('Cannot make topk recommendation with normalized factors')
        # It is possible to skip make recommendation for not-existed keys.
        indexes = self.algo.get_index_pool(keys, group='user')
        keys = [k for k, i in zip(keys, indexes) if i is not None]
        indexes = np.array([i for i in indexes if i is not None], dtype=np.int32)
        if pool is not None:
            pool = self.algo.get_index_pool(pool, group='item')
            if len(pool) == 0:
                raise RuntimeError('pool is empty')
        else:
            # It assume that empty pool menas for all items
            pool = np.array([], dtype=np.int32)
        topks, scores = super()._topk_recommendation_bias(indexes, self.algo.P, self.algo.Q, self.algo.Qb, topk, pool)
        if repr:
            topks = [[self.algo._idmanager.itemids[t] for t in tt if t != -1] for tt in topks]
        return keys, topks, scores


class ParW2V(Parallel):
    def __init__(self, algo, **kwargs):
        num_workers = int(kwargs.get('num_workers', algo.opt.num_workers))
        super().__init__(algo, num_workers=num_workers)

    def most_similar(self, keys, topk=10, pool=None, repr=False, ef_search=-1, use_mmap=True):
        """See the documentation of Parallel."""
        self.algo.normalize(group='item')
        indexes = self.algo.get_index_pool(keys, group='item')
        keys = [k for k, i in zip(keys, indexes) if i is not None]
        indexes = np.array([i for i in indexes if i is not None], dtype=np.int32)
        if pool is not None:
            pool = self.algo.get_index_pool(pool, group='item')
            if len(pool) == 0:
                raise RuntimeError('pool is empty')
        else:
            # It assume that empty pool menas for all items
            pool = np.array([], dtype=np.int32)
        topks, scores = super()._most_similar('item', indexes, self.algo.L0, topk, pool, ef_search, use_mmap)
        if repr:
            mo = np.int32(-1)
            topks = [[self.algo._idmanager.itemids[t] for t in tt if t != mo]
                     for tt in topks]
        return topks, scores

    def topk_recommendation(self, keys, topk=10, pool=None):
        raise NotImplemented


# TODO: Re-think about CFR internal data structure.
class ParCFR(Parallel):
    pass



# -*- coding: utf-8 -*-
"""Buffalo
"""
DOCLINES = __doc__.split("\n")

import os
import sys
import pathlib
import platform
import sysconfig
import subprocess
from setuptools import setup
from cuda_setup import CUDA, build_ext
from distutils.extension import Extension

import n2
import numpy

# TODO: Python3 Support
if sys.version_info[:3] < (3, 6):
    raise RuntimeError("Python version 3.6 or later required.")

assert platform.system() == 'Linux'  # TODO: MacOS
numpy_include_dirs = os.path.split(numpy.__file__)[0] + '/core/include'
n2_shared_object = n2.__file__

MAJOR = 1
MINOR = 0
MICRO = 8
Release = True
STAGE = {True: '', False: 'b'}.get(Release)
VERSION = f'{MAJOR}.{MINOR}.{MICRO}{STAGE}'
STATUS = {False: 'Development Status :: 4 - Beta',
          True: 'Development Status :: 5 - Production/Stable'}

CLASSIFIERS = """{status}
Programming Language :: C++
Programming Language :: Cython
Programming Language :: Python :: 3.6
Operating System :: POSIX :: Linux
Operating System :: Unix
Operating System :: MacOS
License :: OSI Approved :: Apache Software License""".format(status=STATUS.get(Release))
EXTRA_INCLUDE_DIRS = [numpy_include_dirs,
                      '3rd/json11',
                      '3rd/spdlog/include',
                      '3rd/eigen3']
CLIB_DIR = os.path.join(sysconfig.get_path('purelib'), 'buffalo')
LIBRARY_DIRS = [CLIB_DIR]


def get_extend_compile_flags():
    flags = ['-march=native']
    return flags


class CMakeExtension(Extension, object):
    extension_type = 'cmake'

    def __init__(self, name, **kwargs):
        super(CMakeExtension, self).__init__(name, sources=[])


extend_compile_flags = get_extend_compile_flags()

extensions = [
    CMakeExtension(name="cbuffalo"),
    Extension(name="buffalo.algo._als",
              sources=['buffalo/algo/_als.cpp'],
              include_dirs=['./include'] + EXTRA_INCLUDE_DIRS,
              libraries=['gomp', 'cbuffalo'],
              library_dirs=LIBRARY_DIRS,
              runtime_library_dirs=LIBRARY_DIRS,
              extra_compile_args=['-fopenmp', '-std=c++14', '-ggdb', '-O3'] + extend_compile_flags),
    Extension(name="buffalo.algo._cfr",
              sources=['buffalo/algo/_cfr.cpp'],
              include_dirs=['./include'] + EXTRA_INCLUDE_DIRS,
              libraries=['gomp', 'cbuffalo'],
              library_dirs=LIBRARY_DIRS,
              runtime_library_dirs=LIBRARY_DIRS,
              extra_compile_args=['-fopenmp', '-std=c++14', '-ggdb', '-O3'] + extend_compile_flags),
    Extension(name="buffalo.algo._bpr",
              sources=['buffalo/algo/_bpr.cpp'],
              include_dirs=['./include'] + EXTRA_INCLUDE_DIRS,
              libraries=['gomp', 'cbuffalo'],
              library_dirs=LIBRARY_DIRS,
              runtime_library_dirs=LIBRARY_DIRS,
              extra_compile_args=['-fopenmp', '-std=c++14', '-ggdb', '-O3'] + extend_compile_flags),
    Extension(name="buffalo.algo._w2v",
              sources=['buffalo/algo/_w2v.cpp'],
              include_dirs=['./include'] + EXTRA_INCLUDE_DIRS,
              libraries=['gomp', 'cbuffalo'],
              library_dirs=LIBRARY_DIRS,
              runtime_library_dirs=LIBRARY_DIRS,
              extra_compile_args=['-fopenmp', '-std=c++14', '-ggdb', '-O3'] + extend_compile_flags),
    Extension(name="buffalo.misc.log",
              sources=['buffalo/misc/log.cpp'],
              include_dirs=['./include'] + EXTRA_INCLUDE_DIRS,
              libraries=['gomp', 'cbuffalo'],
              library_dirs=LIBRARY_DIRS,
              runtime_library_dirs=LIBRARY_DIRS,
              extra_compile_args=['-fopenmp', '-std=c++14', '-ggdb', '-O3'] + extend_compile_flags),
    Extension(name="buffalo.data.fileio",
              sources=['buffalo/data/fileio.cpp'],
              libraries=['gomp'],
              extra_compile_args=['-fopenmp', '-std=c++14', '-ggdb', '-O3'] + extend_compile_flags),
    Extension(name="buffalo.parallel._core",
              sources=['buffalo/parallel/_core.cpp'],
              libraries=['gomp'],
              include_dirs=EXTRA_INCLUDE_DIRS + ['./3rd/n2/include'],
              library_dirs=LIBRARY_DIRS,
              runtime_library_dirs=LIBRARY_DIRS,
              extra_objects=[n2_shared_object],
              extra_compile_args=['-fopenmp', '-std=c++14', '-ggdb', '-O3'] + extend_compile_flags),
]

if CUDA:
    extra_compile_args = ['-std=c++14', '-ggdb', '-O3'] + extend_compile_flags
    extensions.append(Extension("buffalo.algo.cuda._als",
                                sources=["buffalo/algo/cuda/_als.cpp",
                                         "lib/cuda/als/als.cu",
                                         "./3rd/json11/json11.cpp",
                                         "lib/misc/log.cc"],
                                language="c++",
                                extra_compile_args=extra_compile_args,
                                library_dirs=[CUDA['lib64']],
                                libraries=['cudart', 'cublas', 'curand'],
                                include_dirs=["./include", numpy_include_dirs,
                                              CUDA['include'], "./3rd/json11",
                                              "./3rd/spdlog/include"]))
else:
    print("Failed to find CUDA toolkit. Building without GPU acceleration.")


# Return the git revision as a string
def git_version():
    def _minimal_ext_cmd(cmd):
        # construct minimal environment
        env = {}
        for k in ['SYSTEMROOT', 'PATH']:
            v = os.environ.get(k)
            if v is not None:
                env[k] = v
        out = subprocess.Popen(cmd, stdout=subprocess.PIPE, env=env).communicate()[0]
        return out

    try:
        out = _minimal_ext_cmd(['git', 'rev-parse', 'HEAD'])
        GIT_REVISION = out.strip().decode('ascii')
    except OSError:
        GIT_REVISION = "Unknown"

    return GIT_REVISION


def write_version_py(filename='buffalo/version.py'):
    cnt = """
short_version = '%(version)s'
git_revision = '%(git_revision)s'
"""
    GIT_REVISION = git_version()
    with open(filename, 'w') as fout:
        fout.write(cnt % {'version': VERSION,
                          'git_revision': GIT_REVISION})


class BuildExtension(build_ext, object):
    def run(self):
        for ext in self.extensions:
            if hasattr(ext, 'extension_type') and ext.extension_type == 'cmake':
                self.cmake(ext)
        self.cythonize()
        super(BuildExtension, self).run()

    def cythonize(self):
        ext_files = ['buffalo/algo/_als.pyx',
                     'buffalo/algo/cuda/_als.pyx',
                     'buffalo/algo/_bpr.pyx',
                     'buffalo/algo/_w2v.pyx',
                     'buffalo/misc/log.pyx',
                     'buffalo/algo/_cfr.pyx',
                     'buffalo/parallel/_core.pyx',
                     'buffalo/data/fileio.pyx']
        for path in ext_files:
            from Cython.Build import cythonize
            cythonize(path)

    def cmake(self, ext):
        cwd = pathlib.Path().absolute()

        build_temp = pathlib.Path(self.build_temp)
        build_temp.mkdir(parents=True, exist_ok=True)

        build_type = 'Debug' if self.debug else 'Release'

        cmake_args = [
            '-DCMAKE_BUILD_TYPE=' + build_type,
            '-DCMAKE_LIBRARY_OUTPUT_DIRECTORY=' + CLIB_DIR,
        ]

        build_args = []

        os.chdir(str(build_temp))
        self.spawn(['cmake', str(cwd)] + cmake_args)
        if not self.dry_run:
            self.spawn(['cmake', '--build', '.'] + build_args)
        os.chdir(str(cwd))


def setup_package():
    write_version_py()
    cmdclass = {
        'build_ext': BuildExtension
    }

    build_requires = [l.strip() for l in open('requirements.txt')]

    metadata = dict(
        name='buffalo',
        maintainer="lucas kwangseob kim",
        maintainer_email="recoteam@kakaocorp.com",
        description=DOCLINES[0],
        long_description="\n".join(DOCLINES[2:]),
        url="https://github.com/kakao/buffalo",
        download_url="https://github.com/kakao/buffalo/releases",
        include_package_data=False,
        license='Apache2',
        packages=['buffalo/algo/',
                  'buffalo/algo/cuda',
                  'buffalo/algo/tensorflow',
                  'buffalo/data/',
                  'buffalo/evaluate/',
                  'buffalo/parallel/',
                  'buffalo/misc/',
                  'buffalo/'],
        cmdclass=cmdclass,
        classifiers=[_f for _f in CLASSIFIERS.split('\n') if _f],
        platforms=['Linux', 'Mac OSX', 'Unix'],
        ext_modules=extensions,
        install_requires=build_requires,
        entry_points={
            'console_scripts': [
                'Buffalo = buffalo.cli:_cli_buffalo',
            ]
        },
        python_requires='>=3.6',
    )

    metadata['version'] = VERSION
    setup(**metadata)


if __name__ == '__main__':
    setup_package()



# cython: experimental_cpp_class_def=True, language_level=3
# distutils: language=c++
# -*- coding: utf-8 -*-
import cython
from libcpp cimport bool
from libcpp.string cimport string
from libc.stdint cimport int32_t, int64_t

import numpy as np
cimport numpy as np


cdef extern from "buffalo/algo_impl/bpr/bpr.hpp" namespace "bpr":
    cdef cppclass CBPRMF:
        void release() nogil except +
        bool init(string) nogil except +
        void initialize_model(float*, int32_t,
                              float*, int32_t,
                              float*,
                              int64_t) nogil except +
        void set_cumulative_table(int64_t*, int)
        void launch_workers()
        void add_jobs(int,
                      int,
                      int64_t*,
                      int32_t*) nogil except +
        double compute_loss(int32_t,
                            int32_t*,
                            int32_t*,
                            int32_t*) nogil except +
        double join()
        void wait_until_done()
        void update_parameters()


cdef class CyBPRMF:
    """CBPRMF object holder"""
    cdef CBPRMF* obj  # C-BPRMF object

    def __cinit__(self):
        self.obj = new CBPRMF()

    def __dealloc__(self):
        self.obj.release()
        del self.obj

    def init(self, option_path):
        return self.obj.init(option_path)

    def initialize_model(self,
                         np.ndarray[np.float32_t, ndim=2] P,
                         np.ndarray[np.float32_t, ndim=2] Q,
                         np.ndarray[np.float32_t, ndim=2] Qb,
                         int64_t num_total_samples):
        self.obj.initialize_model(&P[0, 0], P.shape[0],
                                  &Q[0, 0], Q.shape[0],
                                  &Qb[0, 0],
                                  num_total_samples)

    def launch_workers(self):
        self.obj.launch_workers()

    def set_cumulative_table(self,
                             np.ndarray[np.int64_t, ndim=1] cum_table,
                             int cum_table_size):
        self.obj.set_cumulative_table(&cum_table[0], cum_table_size)

    @cython.boundscheck(False)
    @cython.wraparound(False)
    def add_jobs(self,
                 int start_x,
                 int next_x,
                 np.ndarray[np.int64_t, ndim=1] indptr,
                 np.ndarray[np.int32_t, ndim=1] positives):
        self.obj.add_jobs(start_x,
                          next_x,
                          &indptr[0],
                          &positives[0])

    def compute_loss(self,
                     np.ndarray[np.int32_t, ndim=1] users,
                     np.ndarray[np.int32_t, ndim=1] positives,
                     np.ndarray[np.int32_t, ndim=1] negatives):
        return self.obj.compute_loss(users.shape[0],
                                     &users[0],
                                     &positives[0],
                                     &negatives[0])

    def update_parameters(self):
        self.obj.update_parameters()

    def join(self):
        return self.obj.join()

    def wait_until_done(self):
        self.obj.wait_until_done()



# Adapted from https://github.com/rmcgibbo/npcuda-example and
# https://github.com/cupy/cupy/blob/master/cupy_setup_build.py
import logging
import os
import sys

from distutils import ccompiler, errors, msvccompiler, unixccompiler
from setuptools.command.build_ext import build_ext as setuptools_build_ext


def find_in_path(name, path):
    "Find a file in a search path"
    # adapted fom http://code.activestate.com/recipes/52224-find-a-file-given-a-search-path/
    for dir in path.split(os.pathsep):
        binpath = os.path.join(dir, name)
        if os.path.exists(binpath):
            return os.path.abspath(binpath)
    return None


def locate_cuda():
    """Locate the CUDA environment on the system
    If a valid cuda installation is found this returns a dict with keys 'home', 'nvcc', 'include',
    and 'lib64' and values giving the absolute path to each directory.
    Starts by looking for the CUDAHOME env variable. If not found, everything is based on finding
    'nvcc' in the PATH.
    If nvcc can't be found, this returns None
    """
    nvcc_bin = 'nvcc'
    if sys.platform.startswith("win"):
        nvcc_bin = 'nvcc.exe'

    # check env variables CUDA_HOME, CUDAHOME, CUDA_PATH.
    for env_name in ['CUDA_PATH', 'CUDAHOME', 'CUDA_HOME']:
        if env_name not in os.environ:
            continue
        home = os.environ[env_name]
        nvcc = os.path.join(home, 'bin', nvcc_bin)
    else:
        # otherwise, search the PATH for NVCC
        nvcc = find_in_path(nvcc_bin, os.environ['PATH'])
        if nvcc is None:
            logging.warning('The nvcc binary could not be located in your $PATH. Either add it to '
                            'your path, or set $CUDA_HOME to enable CUDA extensions')
            return None
        home = os.path.dirname(os.path.dirname(nvcc))

    cudaconfig = {'home': home,
                  'nvcc': nvcc,
                  'include': os.path.join(home, 'include'),
                  'lib64':   os.path.join(home, 'lib64')}
    post_args = ['-gencode=arch=compute_30,code=sm_30',
                 '-gencode=arch=compute_50,code=sm_50',
                 '-gencode=arch=compute_60,code=sm_60',
                 '-gencode=arch=compute_60,code=compute_60',
                 '--ptxas-options=-v', '-O2']

    if sys.platform == "win32":
        cudaconfig['lib64'] = os.path.join(home, 'lib', 'x64')
        post_args += ['-Xcompiler', '/MD', '-std=c++11']
    else:
        post_args += ['-c', '--compiler-options', "'-fPIC'", '-std=c++11']

    for k, v in cudaconfig.items():
        if not os.path.exists(v):
            logging.warning('The CUDA %s path could not be located in %s', k, v)
            return None

    cudaconfig['post_args'] = post_args
    return cudaconfig


# This code to build .cu extensions with nvcc is taken from cupy:
# https://github.com/cupy/cupy/blob/master/cupy_setup_build.py
class _UnixCCompiler(unixccompiler.UnixCCompiler):
    src_extensions = list(unixccompiler.UnixCCompiler.src_extensions)
    src_extensions.append('.cu')

    def _compile(self, obj, src, ext, cc_args, extra_postargs, pp_opts):
        # For sources other than CUDA C ones, just call the super class method.
        if os.path.splitext(src)[1] != '.cu':
            return unixccompiler.UnixCCompiler._compile(
                self, obj, src, ext, cc_args, extra_postargs, pp_opts)

        # For CUDA C source files, compile them with NVCC.
        _compiler_so = self.compiler_so
        try:
            nvcc_path = CUDA['nvcc']
            post_args = CUDA['post_args']
            # TODO? base_opts = build.get_compiler_base_options()
            self.set_executable('compiler_so', nvcc_path)

            return unixccompiler.UnixCCompiler._compile(
                self, obj, src, ext, cc_args, post_args, pp_opts)
        finally:
            self.compiler_so = _compiler_so


class _MSVCCompiler(msvccompiler.MSVCCompiler):
    _cu_extensions = ['.cu']

    src_extensions = list(unixccompiler.UnixCCompiler.src_extensions)
    src_extensions.extend(_cu_extensions)

    def _compile_cu(self, sources, output_dir=None, macros=None,
                    include_dirs=None, debug=0, extra_preargs=None,
                    extra_postargs=None, depends=None):
        # Compile CUDA C files, mainly derived from UnixCCompiler._compile().
        macros, objects, extra_postargs, pp_opts, _build = \
            self._setup_compile(output_dir, macros, include_dirs, sources,
                                depends, extra_postargs)

        compiler_so = CUDA['nvcc']
        cc_args = self._get_cc_args(pp_opts, debug, extra_preargs)
        post_args = CUDA['post_args']

        for obj in objects:
            try:
                src, _ = _build[obj]
            except KeyError:
                continue
            try:
                self.spawn([compiler_so] + cc_args + [src, '-o', obj] + post_args)
            except errors.DistutilsExecError as e:
                raise errors.CompileError(str(e))

        return objects

    def compile(self, sources, **kwargs):
        # Split CUDA C sources and others.
        cu_sources = []
        other_sources = []
        for source in sources:
            if os.path.splitext(source)[1] == '.cu':
                cu_sources.append(source)
            else:
                other_sources.append(source)

        # Compile source files other than CUDA C ones.
        other_objects = msvccompiler.MSVCCompiler.compile(
            self, other_sources, **kwargs)

        # Compile CUDA C sources.
        cu_objects = self._compile_cu(cu_sources, **kwargs)

        # Return compiled object filenames.
        return other_objects + cu_objects


class cuda_build_ext(setuptools_build_ext):
    """Custom `build_ext` command to include CUDA C source files."""

    def run(self):
        if CUDA is not None:
            def wrap_new_compiler(func):
                def _wrap_new_compiler(*args, **kwargs):
                    try:
                        return func(*args, **kwargs)
                    except errors.DistutilsPlatformError:
                        if not sys.platform == 'win32':
                            CCompiler = _UnixCCompiler
                        else:
                            CCompiler = _MSVCCompiler
                        return CCompiler(
                            None, kwargs['dry_run'], kwargs['force'])
                return _wrap_new_compiler
            ccompiler.new_compiler = wrap_new_compiler(ccompiler.new_compiler)
            # Intentionally causes DistutilsPlatformError in
            # ccompiler.new_compiler() function to hook.
            self.compiler = 'nvidia'

        setuptools_build_ext.run(self)


CUDA = locate_cuda()
build_ext = cuda_build_ext if CUDA else setuptools_build_ext



# -*- coding: utf-8 -*-
import os
import abc
import json
import atexit
import psutil
import warnings
import tempfile
import subprocess

from buffalo.misc import log

_temporary_files = []


class Option(dict):
    def __init__(self, *args, **kwargs):
        import json
        args = [arg if isinstance(arg, dict) else json.loads(open(arg).read())
                for arg in args]
        super(Option, self).__init__(*args, **kwargs)
        for arg in args:
            if isinstance(arg, dict):
                for k, v in arg.items():
                    if isinstance(v, dict):
                        self[k] = Option(v)
                    else:
                        self[k] = v

        if kwargs:
            for k, v in kwargs.items():
                if isinstance(v, dict):
                    self[k] = Option(v)
                else:
                    self[k] = v

    def __getattr__(self, attr):
        return self.get(attr)

    def __setattr__(self, key, value):
        self.__setitem__(key, value)

    def __setitem__(self, key, value):
        super(Option, self).__setitem__(key, value)
        self.__dict__.update({key: value})

    def __delattr__(self, item):
        self.__delitem__(item)

    def __delitem__(self, key):
        super(Option, self).__delitem__(key)
        del self.__dict__[key]

    def __getstate__(self):
        return vars(self)

    def __setstate__(self, state):
        vars(self).update(state)


class InputOptions(abc.ABC):
    def __init__(self, *args, **kwargs):
        pass

    @abc.abstractmethod
    def get_default_option(self) -> dict:
        pass

    def is_valid_option(self, opt) -> bool:
        default_opt = self.get_default_option()
        keys = self.get_default_option()
        for key in keys:
            if key not in opt:
                raise RuntimeError('{} not exists on Option'.format(key))
            expected_type = type(default_opt[key])
            if not isinstance(opt.get(key), expected_type):
                raise RuntimeError('Invalid type for {}, {} expected. '.format(key, type(default_opt[key])))
        return True

    def create_temporary_option_from_dict(self, opt) -> str:
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", ResourceWarning)
            str_opt = json.dumps(opt)
            tmp = tempfile.NamedTemporaryFile(mode='w', dir=opt.get('tmp_dir', '/tmp/'), delete=False)
            tmp.write(str_opt)
            _temporary_files.append(tmp.name)
            return tmp.name


def copy_to_temporary_file(source_path, ignore_lines=0, chunk_size=8192, binary=False):
    W = 'w' if not binary else 'wb'
    R = 'r' if not binary else 'rb'
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", ResourceWarning)
        with tempfile.NamedTemporaryFile(mode=W, delete=False) as w:
            fin = open(source_path, mode=R)
            for _ in range(ignore_lines):
                fin.readline()
            while True:
                chunk = fin.read(chunk_size)
                if chunk:
                    w.write(chunk)
                if len(chunk) != chunk_size:
                    break
            w.close()
            return w.name


def psort(path, parallel=-1, field_seperator=' ', key=1, tmp_dir='/tmp/', buffer_mb=1024, output=None):
    # TODO: We need better way for OS/platform compatibility.
    # we need compatibility checking routine for this method.
    commands = ['sort', '-n', '-s']
    if parallel == -1:
        parallel = psutil.cpu_count()
    if parallel > 0:
        commands.extend(['--parallel', parallel])
    if not output:
        output = path
    commands.extend(['-t', '{}'.format(field_seperator)])
    commands.extend(['-k', key])
    commands.extend(['-T', tmp_dir])
    commands.extend(['-S', '%sM' % buffer_mb])
    commands.extend(['-o', output])
    commands.append(path)
    try:
        subprocess.check_output(map(str, commands), stderr=subprocess.STDOUT, env={'LC_ALL': 'C'})
    except Exception as e:
        log.get_logger().error('Unexpected error: %s for %s' % (str(e), ' '.join(list(map(str, commands)))))
        raise


def get_temporary_file(root='/tmp/', write_mode='w'):
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", ResourceWarning)
        w = tempfile.NamedTemporaryFile(mode=write_mode, dir=root, delete=False)
        _temporary_files.append(w.name)
        return w.name


@atexit.register
def __cleanup_tempory_files():
    for path in _temporary_files:
        if os.path.isfile(path):
            os.remove(path)


def register_cleanup_file(path):
    _temporary_files.append(path)
